package prt;

import addr.addrMac;
import cfg.cfgInit;
import ifc.ifcDn;
import ifc.ifcThread;
import ifc.ifcUp;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.List;
import pack.packHolder;
import user.userFormat;
import user.userFlash;
import util.bits;
import util.counter;
import util.debugger;
import util.logger;
import util.notifier;
import util.state;
import util.version;

/**
 * redundancy protocol
 *
 * @author matecsaba
 */
public class prtRedun implements Runnable {

    /**
     * speaking
     */
    public final static int statSpeak = 1;

    /**
     * standby
     */
    public final static int statStandby = 2;

    /**
     * active
     */
    public final static int statActive = 3;

    /**
     * keepalive time
     */
    public final static int timeKeep = 200;

    /**
     * time multiplier
     */
    public final static int timeMult = 5;

    /**
     * hold time
     */
    public final static int timeHold = timeKeep * timeMult;

    /**
     * my magic number
     */
    protected static int magic = 0;

    /**
     * current state
     */
    protected static int state = 0;

    private final static List<prtRedunIfc> ifaces = new ArrayList<prtRedunIfc>();

    public void run() {
        try {
            for (;;) {
                bits.sleep(timeKeep);
                if (!cfgInit.booting) {
                    state = statActive;
                }
                for (int i = 0; i < ifaces.size(); i++) {
                    ifaces.get(i).doPack(prtRedunPck.typHello, new packHolder(true, true));
                }
            }
        } catch (Exception e) {
            logger.exception(e);
        }
    }

    /**
     * generate show output
     *
     * @return output
     */
    public static userFormat doShow() {
        userFormat l = new userFormat("|", "iface|state|magic");
        l.add("self|" + state + "|" + magic);
        for (int i = 0; i < ifaces.size(); i++) {
            prtRedunIfc ifc = ifaces.get(i);
            l.add(ifc.doShow());
        }
        return l;
    }

    /**
     * add one physical interface
     *
     * @param name name of interface
     * @param thrd interface thread handler
     */
    public static void ifcAdd(String name, ifcThread thrd) {
        prtRedunIfc ifc = new prtRedunIfc();
        ifc.doInit(name, thrd);
        ifaces.add(ifc);
    }

    /**
     * find best peer
     *
     * @return best peer index, -1 if none
     */
    public static int findActive() {
        long tim = bits.getTime();
        int bi = -1;
        int bm = magic;
        for (int i = 0; i < ifaces.size(); i++) {
            prtRedunIfc ifc = ifaces.get(i);
            if (!ifc.bidir) {
                continue;
            }
            if ((tim - ifc.time) > timeHold) {
                continue;
            }
            if (ifc.state == statActive) {
                return i;
            }
            if (ifc.magic < bm) {
                continue;
            }
            bi = i;
            bm = ifc.magic;
        }
        return bi;
    }

    /**
     * terminate the redundancy
     */
    public static void doShut() {
    }

    /**
     * initialize the redundancy
     */
    public static void doInit() {
        magic = (bits.randomD() & 0x3fffffff) + 1;
        if (ifaces.size() < 1) {
            state = statActive;
            return;
        }
        state = statSpeak;
        new Thread(new prtRedun()).start();
        bits.sleep(timeHold);
        if (findActive() < 0) {
            return;
        }
        state = statStandby;
        logger.info("initialized as standby");
        for (;;) {
            bits.sleep(timeKeep);
            if (findActive() < 0) {
                break;
            }
        }
        state = statActive;
    }

    /**
     * notify peers to reload
     */
    public static void doNotify() {
        for (int i = 0; i < ifaces.size(); i++) {
            prtRedunIfc ifc = ifaces.get(i);
            ifc.doFile(cfgInit.cfgFileSw);
            ifc.doPack(prtRedunPck.typReload, new packHolder(true, true));
        }
    }

}

class prtRedunPck {

    public static final int ethtyp = 0x8087;

    public static final int size = 12;

    public static final int typHello = 1;

    public static final int typReload = 2;

    public static final int typAck = 3;

    public static final int typFilBeg = 4;

    public static final int typFilDat = 5;

    public static final int typFilEnd = 6;

    public static final String fnBegin = "//";

    public static final String fnStart = fnBegin + "startup";

    public static final String fnCore = fnBegin + "rtr+jar";

    public static final int dataMax = 1024;

    public int type;

    public int state;

    public int magic;

    public int peer;

    public static String typ2str(int i) {
        switch (i) {
            case typHello:
                return "hello";
            case typReload:
                return "reload";
            case typAck:
                return "ack";
            case typFilBeg:
                return "filBeg";
            case typFilDat:
                return "filDat";
            case typFilEnd:
                return "filEnd";
            default:
                return "unknown=" + i;
        }
    }

    public boolean parseHeader(packHolder pck) {
        if (pck.msbGetW(0) != ethtyp) {
            return true;
        }
        type = pck.getByte(2);
        state = pck.getByte(3);
        magic = pck.msbGetD(4);
        peer = pck.msbGetD(8);
        pck.getSkip(size);
        return false;
    }

    public void createHeader(packHolder pck) {
        pck.msbPutW(0, ethtyp);
        pck.putByte(2, type);
        pck.putByte(3, state);
        pck.msbPutD(4, magic);
        pck.msbPutD(8, peer);
        pck.putSkip(size);
        pck.merge2beg();
    }

    public String toString() {
        return "type=" + typ2str(type) + " state=" + state + " magic=" + magic + " peer=" + peer;
    }

}

class prtRedunIfc implements ifcUp {

    private ifcThread lower;

    private counter cntr = new counter();

    private addrMac hwaddr;

    private RandomAccessFile filRx;

    public String name;

    public boolean bidir;

    public int magic;

    public int state;

    public long time;

    public notifier notif = new notifier();

    public int ackRx;

    public void doInit(String nam, ifcThread thrd) {
        name = nam;
        lower = thrd;
        state = 0;
        magic = 0;
        time = 0;
        lower.setFilter(false);
        lower.setUpper(this);
        lower.startLoop();
        hwaddr = (addrMac) lower.getHwAddr();
    }

    public String doShow() {
        return name + "|" + state + "|" + magic;
    }

    public void setParent(ifcDn parent) {
    }

    public void setState(state.states stat) {
    }

    public void closeUp() {
    }

    public counter getCounter() {
        return cntr;
    }

    public void recvPack(packHolder pck) {
        prtRedunPck pckP = new prtRedunPck();
        if (pckP.parseHeader(pck)) {
            return;
        }
        if (debugger.prtRedun) {
            logger.debug("rx " + pckP);
        }
        magic = pckP.magic;
        state = pckP.state;
        bidir = pckP.peer == prtRedun.magic;
        time = bits.getTime();
        switch (pckP.type) {
            case prtRedunPck.typHello:
                if ((magic == prtRedun.magic) && (state >= prtRedun.state)) {
                    cfgInit.stopRouter(true, 6, "magic collision");
                }
                if ((prtRedun.state == prtRedun.statActive) && (state == prtRedun.statActive) && (magic <= prtRedun.magic)) {
                    cfgInit.stopRouter(true, 6, "dual active");
                }
                break;
            case prtRedunPck.typReload:
                if (prtRedun.state == prtRedun.statActive) {
                    break;
                }
                cfgInit.stopRouter(true, 6, "peer request");
                break;
            case prtRedunPck.typAck:
                ackRx = pck.msbGetD(0);
                notif.wakeup();
                break;
            case prtRedunPck.typFilBeg:
                try {
                    filRx.close();
                } catch (Exception e) {
                }
                try {
                    filRx = new RandomAccessFile(version.getFileName() + ".tmp", "rw");
                    filRx.seek(0);
                    filRx.setLength(0);
                } catch (Exception e) {
                }
                doAck(-2);
                break;
            case prtRedunPck.typFilDat:
                int i = pck.msbGetD(0);
                int o = pck.msbGetW(4);
                pck.getSkip(6);
                byte[] buf = new byte[o];
                pck.getCopy(buf, 0, 0, o);
                try {
                    filRx.seek(i);
                    filRx.write(buf);
                } catch (Exception e) {
                }
                doAck(i);
                break;
            case prtRedunPck.typFilEnd:
                try {
                    filRx.close();
                } catch (Exception e) {
                }
                filRx = null;
                String a = pck.getAsciiZ(0, prtRedunPck.dataMax, 0);
                String b = null;
                if (a.equals(prtRedunPck.fnCore)) {
                    b = version.getFileName();
                }
                if (a.equals(prtRedunPck.fnStart)) {
                    b = cfgInit.cfgFileSw;
                }
                a = version.getFileName() + ".tmp";
                userFlash.rename(a, b, true, true);
                doAck(-3);
                break;
        }
    }

    public void doPack(int typ, packHolder pckB) {
        pckB.merge2beg();
        prtRedunPck pckP = new prtRedunPck();
        pckP.type = typ;
        pckP.state = prtRedun.state;
        pckP.magic = prtRedun.magic;
        pckP.peer = magic;
        pckP.createHeader(pckB);
        if (debugger.prtRedun) {
            logger.debug("tx " + pckP);
        }
        pckB.ETHsrc.setAddr(hwaddr);
        pckB.ETHtrg.setAddr(addrMac.getBroadcast());
        lower.sendPack(pckB);
    }

    public void doAck(int ofs) {
        packHolder pck = new packHolder(true, true);
        pck.msbPutD(0, ofs);
        pck.putSkip(4);
        doPack(prtRedunPck.typAck, pck);
    }

    public boolean doRetry(int typ, packHolder pck) {
        ackRx = -1;
        for (int i = 0; i < prtRedun.timeMult; i++) {
            doPack(typ, pck.copyBytes(true, true));
            notif.sleep(prtRedun.timeKeep);
            if (ackRx != -1) {
                return false;
            }
        }
        return true;
    }

    public boolean doFile(String fn) {
        RandomAccessFile fr;
        long siz;
        try {
            fr = new RandomAccessFile(fn, "r");
            siz = fr.length();
        } catch (Exception e) {
            return true;
        }
        if (doRetry(prtRedunPck.typFilBeg, new packHolder(true, true))) {
            try {
                fr.close();
            } catch (Exception e) {
            }
            return true;
        }
        long pos = 0;
        for (; pos < siz;) {
            long rndl = siz - pos;
            if (rndl > prtRedunPck.dataMax) {
                rndl = prtRedunPck.dataMax;
            }
            int rndi = (int) rndl;
            byte buf[] = new byte[rndi];
            try {
                fr.read(buf, 0, rndi);
            } catch (Exception e) {
                return true;
            }
            packHolder pck = new packHolder(true, true);
            pck.msbPutD(0, (int) pos);
            pck.msbPutW(4, buf.length);
            pck.putSkip(6);
            pos += rndl;
            pck.putCopy(buf, 0, 0, buf.length);
            pck.putSkip(buf.length);
            if (doRetry(prtRedunPck.typFilDat, pck)) {
                try {
                    fr.close();
                } catch (Exception e) {
                }
                return true;
            }
        }
        try {
            fr.close();
        } catch (Exception e) {
            return true;
        }
        packHolder pck = new packHolder(true, true);
        pck.putAsciiZ(0, prtRedunPck.dataMax, prtRedunPck.fnStart, 0);
        pck.putSkip(prtRedunPck.dataMax);
        if (doRetry(prtRedunPck.typFilEnd, pck)) {
            return true;
        }
        return false;
    }

}
